from setuptools import setup, find_packages

setup(
    name='envinstall',
    version='',
    packages=find_packages(),
    url='',
    license='',
    author='patrick',
    author_email='',
    description=''
)
